@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2011-2012 Yubico AB
+ * Copyright (c) 2011-2013 Yubico AB
  * All rights reserved.
  *
  * Author : Fredrik Thulin <fredrik@yubico.com>
@@ -56,6 +56,7 @@ const char *usage =
   "\t-2           Send challenge to slot 2.\n"
   "\t-A action    What to do.\n"
   "\t-p path      Specify an output path for the challenge file.\n"
+  "\t-i iters     Number of iterations to use for pbkdf2 (defaults to 10000)\n"
   "\n"
   "\t-v           verbose\n"
   "\t-h           help (this text)\n"
@@ -66,10 +67,10 @@ const char *usage =
   "\n"
   "\n"
   ;
-const char *optstring = "12A:p:vh";
+const char *optstring = "12A:p:i:vh";
 
 static void
-report_yk_error()
+report_yk_error(void)
 {
   if (ykp_errno)
     fprintf(stderr, "Yubikey personalization error: %s\n",
@@ -85,10 +86,11 @@ report_yk_error()
   }
 }
 
-int
+static int
 parse_args(int argc, char **argv,
 	   int *slot, bool *verbose,
 	   char **action, char **output_dir,
+     unsigned int *iterations,
 	   int *exit_code)
 {
   int c;
@@ -107,6 +109,14 @@ parse_args(int argc, char **argv,
     case 'p':
       *output_dir = optarg;
       break;
+    case 'i':
+      *iterations = strtoul(optarg, NULL, 10);
+      if(*iterations == 0) {
+        fprintf(stderr, "iterations must be numeric, %s isn't.\n", optarg);
+        *exit_code = 1;
+        return 0;
+      }
+      break;
     case 'v':
       *verbose = true;
       break;
@@ -121,20 +131,19 @@ parse_args(int argc, char **argv,
   return 1;
 }
 
-int
-do_add_hmac_chalresp(YK_KEY *yk, uint8_t slot, bool verbose, char *output_dir, int *exit_code)
+static int
+do_add_hmac_chalresp(YK_KEY *yk, uint8_t slot, bool verbose, char *output_dir, unsigned int iterations, int *exit_code)
 {
   char buf[CR_RESPONSE_SIZE + 16];
   CR_STATE state;
-  unsigned int flags = 0;
   int ret = 0;
   unsigned int response_len;
   char *fn;
   struct passwd *p;
   FILE *f = NULL;
 
+  state.iterations = iterations;
   state.slot = slot;
-  flags |= YK_FLAG_MAYBLOCK;
   *exit_code = 1;
 
   p = getpwuid (getuid ());
@@ -149,10 +158,9 @@ do_add_hmac_chalresp(YK_KEY *yk, uint8_t slot, bool verbose, char *output_dir, i
   */
   
   if (!output_dir){
-      const char *pathname = p->pw_dir; 
+      struct stat st;
       char fullpath[256];
       snprintf(fullpath, 256,"%s/.yubico",p->pw_dir);
-      struct stat st;
       
       //check if directory exists     
       if (stat(fullpath,&st)!=0 ){     
@@ -185,7 +193,7 @@ do_add_hmac_chalresp(YK_KEY *yk, uint8_t slot, bool verbose, char *output_dir, i
   state.challenge_len = CR_CHALLENGE_SIZE;
 
   if (! challenge_response(yk, state.slot, state.challenge, CR_CHALLENGE_SIZE,
-			   true, flags, verbose,
+			   true, true, verbose,
 			   buf, sizeof(buf), &response_len))
     goto out;
 
@@ -195,14 +203,13 @@ do_add_hmac_chalresp(YK_KEY *yk, uint8_t slot, bool verbose, char *output_dir, i
   {
     char buf2[CR_RESPONSE_SIZE + 16];
     char challenge[CR_CHALLENGE_SIZE];
-    CR_STATE state2;
 
     if (generate_random(challenge, CR_CHALLENGE_SIZE)) {
       fprintf (stderr, "FAILED getting %i bytes of random data\n", CR_CHALLENGE_SIZE);
       goto out;
     }
     if (! challenge_response(yk, state.slot, challenge, CR_CHALLENGE_SIZE,
-          true, flags, verbose,
+          true, true, verbose,
           buf2, sizeof(buf2), &response_len))
       goto out;
 
@@ -253,6 +260,7 @@ main(int argc, char **argv)
   char *action = ACTION_ADD_HMAC_CHALRESP;
   char *output_dir = NULL;
   int slot = 1;
+  unsigned int iterations = CR_DEFAULT_ITERATIONS;
 
   ykp_errno = 0;
   yk_errno = 0;
@@ -260,7 +268,7 @@ main(int argc, char **argv)
   if (! parse_args(argc, argv,
 		   &slot, &verbose,
 		   &action, &output_dir,
-		   &exit_code))
+       &iterations, &exit_code))
     goto err;
 
   exit_code = 1;
@@ -275,7 +283,7 @@ main(int argc, char **argv)
     if (! check_firmware_version(yk, verbose, false))
       goto err;    
 
-    if (! do_add_hmac_chalresp (yk, slot, verbose, output_dir, &exit_code))
+    if (! do_add_hmac_chalresp (yk, slot, verbose, output_dir, iterations, &exit_code))
       goto err;
   } else {
     fprintf (stderr, "Unknown action '%s'\n", action);
